/*
 * AlphaBeta - Minimax - Simulation v0.1
 * 
 */


import java.lang.Thread;
import java.applet.Applet;
import java.awt.*;


/**
 * the main game class. represents the board with alphaBetaAlgo
 * 
 */
class gameTicTacToeBoard {

    // some useful statics
    public static final int EMPTY = 0;
    public static final int X = 1;
    public static final int O = 2;
    public static final int DRAW = 3;
	public static final int MIN = 1;
	public static final int MAX = 2;
	public static long glb_operation_count = 0;

    // helper for toString method
    public static char[] to_string = {'#','X','O'};

    // board
    private int[][] board;

    private int best_i;
    private int best_j;

    /*
     * each winning line gets a bit
     */
    private static int[][] board_mask = {
        {1|8|64,2|8,4|8|128},
        {1|16,2|16|64|128,4|16},
        {1|32|128,2|32,4|32|64},
    };

    /*
     * each winning line gets an unset bit, all else
     * is set. (makes for easier masking later on)
     */
    static {
        for(int i=0;i<3;i++)
            for(int j=0;j<3;j++)
                board_mask[i][j] = ~board_mask[i][j];
    }

    /**
     * default constructor; allocate mem, and
     * init the board to empty
     */
    public gameTicTacToeBoard(){
        board = new int[3][3];
        for(int i=0;i<3;i++)
            for(int j=0;j<3;j++)
                board[i][j] = EMPTY;
    }

    /**
     * copy constructor; allocate mem, and
     * init the board to the passed board.
     */
    public gameTicTacToeBoard(gameTicTacToeBoard b){
        board = new int[3][3];
        for(int i=0;i<3;i++)
            for(int j=0;j<3;j++)
                board[i][j] = b.board[i][j];
    }

    /**
     * get number of winning lines for a specific player
     *
     * @param lines Either gameTicTacToeBoard.X or gameTicTacToeBoard.O
     * @return number of possible winning lines for l
     */
    public int getWinLinesForPlayer(int lines){
        int masked = 0xFF;
        int otherplayer = lines == X ? O:X;
        for(int i=0;i<3;i++)
            for(int j=0;j<3;j++)
                if(board[i][j] == otherplayer)
                    masked &= board_mask[i][j]; //mask out win lines
        int c=0;
        for(int i=1;i<0x100;i<<=1)  //count win lines left
            if((masked & i) != 0)
                c++;
        return c;
    }

    /**
     * evaluate board to a specific player
     * (this is just a diff of player's open win lines)
     */
    public int getBoardEval(int l){
        return getWinLinesForPlayer(l) -
            getWinLinesForPlayer(l == X ? O:X);
    }

    /**
     * set value at a specific board location
     */
    public void setValOnBoard(int row,int col,int val){
        board[row][col] = val;
    }

    /**
     * get value at a specific board location.
     */
    public int getValOnBoard(int row,int col){
        return board[row][col];
    }

    // enable quicker end-game check
    // that's fucking great math stuff
    private static int[][][] wingamechecks = {
        { {0,0},{0,1},{0,2} },
        { {0,0},{1,0},{2,0} },
        { {0,0},{1,1},{2,2} },
        { {0,1},{1,1},{2,1} },
        { {0,2},{1,2},{2,2} },
        { {1,0},{1,1},{1,2} },
        { {2,0},{2,1},{2,2} },
        { {2,0},{1,1},{0,2} }
    };


    /**
     * see if game is over. returns 0 if not, else returns
     * winner, or DRAW if game ended in draw.
     */
    public int getGameWinner(){
        // check for actual winner
        for(int i=0;i<8;i++){
            int a = board[wingamechecks[i][0][0]][wingamechecks[i][0][1]];
            int b = board[wingamechecks[i][1][0]][wingamechecks[i][1][1]];
            int c = board[wingamechecks[i][2][0]][wingamechecks[i][2][1]];
            if(a != EMPTY && a == b && b == c){
                return a;
            }
        }
        // no winner, but maybe a draw
        for(int i=0;i<3;i++)
            for(int j=0;j<3;j++)
                if(board[i][j] == EMPTY)
                    return 0;
        // draw ... (no spaces left for a move)
        return DRAW;
    }

    /**
     * alpha beta procedure
     */
    // --------- APLHABETACUT-ALGO --------- START ---------
    public static int alphaBeta(gameTicTacToeBoard board, int alpha, int beta,  int depth, int max_depth){
        int winner = board.getGameWinner();
        glb_operation_count++;
        if(winner != 0){
            if(winner == DRAW){
                alpha = 0; 	
            }else{
                alpha = winner == depth ? 100:-100;
            }
//          [DEBUG] System.out.println("winner spotted:"+winner+"; alpha:"+alpha+"; max_depth:"+max_depth);
        }else if(max_depth <= 0){
            alpha = board.getBoardEval(depth);
        }else{
            board.best_i = -1;
            board.best_j = -1;

            // for all children
            L1: for(int i=0;i<3;i++){
            	//glb_operation_count++;	// for
                for(int j=0;j<3;j++){
                	//glb_operation_count++; // for
                    if(board.getValOnBoard(i, j) == EMPTY){
                        // make move, and recursively eval it
                        gameTicTacToeBoard moveboard = new gameTicTacToeBoard(board);
                        moveboard.setValOnBoard(i, j, depth);
                        
                        // our negated beta becomes the other guy's alpha, etc.
                        int val = alphaBeta(moveboard, -beta, -alpha, depth == X ? O:X, max_depth-1);
                        
                        // make sure that we always have a move...
                        if(board.best_i == -1){
                            board.best_i = i;
                            board.best_j = j;
                        }

                        // is this a better move?
                        if(val > alpha){
                            alpha = val;
                            board.best_i = i;
                            board.best_j = j;
                        }
                        if(alpha >= beta)
                            break L1;  // cutoff
                    }
                }
            }
        }
        //glb_operation_count++; // return
        return -alpha;
    }
    // --------- APLHABETACUT-ALGO --------- END ---------
  
    
    
    
    // --------- MINIMAX-ALGO --------- START ---------
    public static int miniMax(gameTicTacToeBoard board, int player, int max_depth){
    	glb_operation_count++;
    	// the startcall from makeMove is for the computer (depth = O = 2) as MAX!
    	int best_val;
    	int winner = board.getGameWinner();
        if(winner != 0){
            if(winner == DRAW){
            	best_val = 0;
            }else{
            	best_val = winner == player ? 100:-100;
            }
//          [DEBUG] System.out.println("winner spotted:"+winner+"; alpha:"+alpha+"; max_depth:"+max_depth);
        }else if(max_depth <= 0){
        	best_val = board.getBoardEval(player);
        }else{
            board.best_i = -1;
            board.best_j = -1;
            	
            // MIN
            if (player == MIN)
        	{
        		int minValue = Integer.MAX_VALUE;
        		// for all children
                L1: for(int i=0;i<3;i++){
                	//glb_operation_count++;	// for
                    for(int j=0;j<3;j++){
                    	//glb_operation_count++; // for
    	                    if(board.getValOnBoard(i, j) == EMPTY){
    	                        // make move, and recursively eval it
    	                        gameTicTacToeBoard moveboard = new gameTicTacToeBoard(board);
    	                        moveboard.setValOnBoard(i, j, player);
    	                        // now step into the level of max
    	                        int val = miniMax(moveboard, MAX, max_depth-1);
    	                        //minValue = Math.min(val,minValue);
    	                        //is this a better move?
    	                        if(val < minValue){
    	                        	minValue = val;
    	                            board.best_i = i;
    	                            board.best_j = j;
    	                        }
    	                        // make sure that we always have a move...
    	                        if(board.best_i == -1){
    	                            board.best_i = i;
    	                            board.best_j = j;
    	                        }
    	                    }
                    	}
                }
        		best_val = minValue;
        	}
        	else // (player == MAX)
            {
        		int maxValue = -Integer.MAX_VALUE;
        		// for all children
                L2: for(int i=0;i<3;i++){
                	//glb_operation_count++;	// for
                    for(int j=0;j<3;j++){
                    	//glb_operation_count++; // for
    	                    if(board.getValOnBoard(i, j) == EMPTY){
    	                        // make move, and recursively eval it
    	                        gameTicTacToeBoard moveboard = new gameTicTacToeBoard(board);
    	                        moveboard.setValOnBoard(i, j, player);
    	                        // now step into the level of min
    	                        int val = miniMax(moveboard, MIN, max_depth-1);
    	                        //maxValue = Math.max(val,maxValue);
    	                        //is this a better move?
    	                        if(val > maxValue){
    	                        	maxValue = val;
    	                            board.best_i = i;
    	                            board.best_j = j;
    	                        }
    	                        
    	                        // make sure that we always have a move...
    	                        if(board.best_i == -1){
    	                            board.best_i = i;
    	                            board.best_j = j;
    	                        }
    	                    }
                    	}
                }        		
        		best_val = maxValue;
            }            
        }
        //glb_operation_count++; // return
        return best_val;
    }
    // --------- MINIMAX-ALGO --------- END ---------
    
    
    
    public int makeMove(int lines,int skill){
    	int val;
    	glb_operation_count = 0;
    	if(1==1){
    		// use ALPHA-BETA Algo
    		val = alphaBeta(this,-100,100,lines,skill);
    		System.out.println("Algo: AlphaBeta - Skill: " + skill + " - Laufzeit: "+ glb_operation_count);
    	}else{
    		// use MINIMAX Algo
    		//does not work complete 
    		val = miniMax(this,lines,skill);
    		System.out.println("Algo: MiniMax - Skill: " + skill + " - Laufzeit: "+ glb_operation_count);
    	}
//        [DEBUG] System.out.println("val: "+val+" ("+best_i+","+best_j+")");
    	
        setValOnBoard(best_i,best_j,lines);
        return getGameWinner();
    }

    // virtual board used for simulation
    /**
     * return string representation of the board
     */
    public String toString(){
        StringBuffer s = new StringBuffer();
        for(int i=0;i<3;i++){
            for(int j=0;j<3;j++)
                s.append(to_string[board[i][j]]);
            s.append('\n');
        }
        return s.toString();
    }

}



// --------- THE APPLET SECTION ---------
/**
 * Applet helper class. doesn't have any real logic in it,
 * (leaving all the game stuff to the board class).
 * Provides UI for the game.
 *
 * <p>August 15th, 2001 - Fixed image download problem. Images
 * would not be donwloaded until they were being displayed. On
 * a slow connection, there was a noticable delay abetween the
 * time you clicked a mouse, and the time an image would appear.
 * Fix involved invoking a MediaTracker to download the images
 * in a seperate thread.
 *
 * 
 */
public class gameTicTacToe extends Applet implements Runnable {

    private Image imageX,imageO,imageXWins,imageOWins,imageXODraw;
    private gameTicTacToeBoard board;

    // track loading of images
    private MediaTracker tracker;

    private Thread m_thread;
    
    // initialize with a higher skill
    private int skill;

    public void init(){

        // initialize media tracker
        tracker = new MediaTracker(this);

        setBackground(Color.white);
        imageO = getImage(getCodeBase(), "o_player.gif");
        imageX = getImage(getCodeBase(), "x_player.gif");
        imageXWins = getImage(getCodeBase(),"x_wins.gif");
        imageOWins = getImage(getCodeBase(),"o_wins.gif");
        imageXODraw = getImage(getCodeBase(),"drawn.gif");

        // track image
        tracker.addImage(imageO,0);
        tracker.addImage(imageX,1);
        tracker.addImage(imageXWins,2);
        tracker.addImage(imageOWins,3);
        tracker.addImage(imageXODraw,4);

        board = new gameTicTacToeBoard();

        try{
            // try to get skill level from html
            skill = Integer.parseInt(getParameter("skill"));
        }catch(Throwable e){
            System.out.println(e.toString());
            // if no param is given set it to a higher skill again
            skill = 9;
        }
    }

    /**
     * start image download thread
     */
    public void start(){
        m_thread = new Thread(this);
        m_thread.start();
    }

    public void stop(){
        m_thread = null;
    }

    /**
     * download all images while the player makes a move...
     */
    public void run(){
        try {
            tracker.waitForAll();
        } catch (InterruptedException e) {
        }
    }

    public void paint(Graphics g) {
    	// GIF-Size is 40px
    	g.setColor(Color.black);
        Dimension d = getSize(); // former <size()> is deprecated
        
        // do some calculations on image offsets
        int xoff = d.width / 3;
        int yoff = d.height / 3;
        
        int gamepic_x = 40;
        int gamepic_y = 40;
        int infopic_x = 120;
        int infopic_y = 60;
        
        int gamepic_posx = (d.width / 6) - gamepic_x / 2; 
        int gamepic_posy = (d.height / 6) - gamepic_y / 2;
        int infopic_posx = (d.width / 2) - infopic_x /2 ;
        int infopic_posy = (d.height / 2) - infopic_y /2 ;
        
        g.drawLine(xoff, 0, xoff, d.height);
        //System.out.println("[DEBUG] x.Anfang " + xoff + " | y.Anf " + 0 + "\nx.Ende " + xoff + " | y.Ende " + d.height);
        g.drawLine(2*xoff, 0, 2*xoff, d.height);
        g.drawLine(0, yoff, d.width, yoff);
        g.drawLine(0, 2*yoff, d.width, 2*yoff);
        for (int r = 0 ; r < 3 ; r++) {
            for (int c = 0 ; c < 3 ; c++) {
                int t = board.getValOnBoard(r,c);
                if (t == gameTicTacToeBoard.O){
                    g.drawImage(imageO, c*xoff + gamepic_posx, r*yoff + gamepic_posy, this);
                    // [DEBUG] System.out.println("For Player o:\n---\nx: " + (c*xoff) + "\ny: " + (c*yoff));
                } else if (t == gameTicTacToeBoard.X) {
                	g.drawImage(imageX, c*xoff + gamepic_posx, r*yoff + gamepic_posy, this);
                	// [DEBUG] System.out.println("For Player x:\n---\nx: " + (c*xoff) + "\ny: " + (c*yoff));
                }
            }
        }
        int winner = board.getGameWinner();
        if(winner != 0){
            if(winner == gameTicTacToeBoard.DRAW){
                g.drawImage(imageXODraw,infopic_posx,infopic_posy, this);
            }else if(winner == gameTicTacToeBoard.X){
                g.drawImage(imageXWins,infopic_posx,infopic_posy, this);
            }else if(winner == gameTicTacToeBoard.O){
                g.drawImage(imageOWins,infopic_posx,infopic_posy, this);
            }
        }
    }


    public boolean mouseUp(Event evt, int x, int y) {
        // if there was a winner...
        if(board.getGameWinner() != 0){
            board = new gameTicTacToeBoard();
            repaint();
        }else{
            Dimension d = getSize();
            // Figure out the row/column
            int c = (x * 3) / d.width;
            int r = (y * 3) / d.height;
            if(board.getValOnBoard (r,c) == gameTicTacToeBoard.EMPTY){
                board.setValOnBoard (r,c,gameTicTacToeBoard.X);
                repaint();
                int winner = board.getGameWinner();
                // if there was a winner...
                if(winner == 0){
                    board.makeMove(gameTicTacToeBoard.O,skill);
                    repaint();
                }
            }
        }
        return true;
    }

}

